"""
### Clip vectors to other geometries ###

Clip vector files with other geometries. Can come from rasters or vectors.
"""

# Standard library
import sys; sys.path.append("../../")

# External
from osgeo import ogr

# Internal
from buteo.utils import gdal_utils, core_utils
from buteo.vector import core_vector


def _buffer_vector(vector, distance, out_path=None):
    """ Internal. """
    if out_path is None:
        out_path = gdal_utils.create_memory_path(
            gdal_utils.get_path_from_dataset(vector),
            prefix="",
            suffix="_buffer",
            add_uuid=True,
        )

    assert core_utils.is_valid_output_path(out_path), "Invalid output path"

    read = core_vector.open_vector(vector)

    driver = ogr.GetDriverByName(gdal_utils.path_to_driver_vector(out_path))
    destination = driver.CreateDataSource(out_path)

    vector_metadata = core_vector._vector_to_metadata(read)

    for layer_idx in range(0, len(vector_metadata["layers"])):
        layer_name = vector_metadata["layers"][layer_idx]["layer_name"]
        vector_layer = read.GetLayer(layer_name)
        destination.CopyLayer(vector_layer, layer_name, ["OVERWRITE=YES"])

        field_names = vector_metadata["layers"][layer_idx]["field_names"]
        if isinstance(distance, str):
            if distance not in field_names:
                raise ValueError(f"Attribute to buffer by not in vector field names: {distance} not in {field_names}")

            sql = f"update {layer_name} set geom=ST_Buffer(geom, {layer_name}.{distance})"
        else:
            sql = f"update {layer_name} set geom=ST_Buffer(geom, {distance})"

        destination.ExecuteSQL(sql, dialect="SQLITE")

    if destination is None:
        raise Exception("Error while running intersect.")

    destination.FlushCache()

    return out_path


def buffer_vector(
    vector,
    distance,
    out_path=None,
    *,
    prefix="",
    suffix="",
    add_uuid=False,
    allow_lists=True,
    overwrite=True,
):
    """
    Buffers a vector with a fixed distance or an attribute.

    ## Args:
    `vector` (_str_/_ogr.DataSource_/_list_): Vector(s) to buffer. </br>
    `distance` (_int_/_float_/_str_): The distance to buffer with. If string, uses the attribute of that name. </br>

    ## Kwargs:
    `out_path` (_str_/_None_): Output path. If None, memory vectors are created. (Default: **None**) </br>
    `prefix` (_str_): Prefix to add to the output path. (Default: **""**) </br>
    `suffix` (_str_): Suffix to add to the output path. (Default: **""**) </br>
    `add_uuid` (_bool_): Add UUID to the output path. (Default: **False**) </br>
    `allow_lists` (_bool_): Allow lists of vectors as input. (Default: **True**) </br>
    `overwrite` (_bool_): Overwrite output if it already exists. (Default: **True**) </br>

    ## Returns:
    (_str_/_list_): Output path(s) of clipped vector(s).
    """
    core_utils.type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    core_utils.type_check(distance, [int, float, str], "distance")
    core_utils.type_check(out_path, [str, None], "out_path")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "suffix")
    core_utils.type_check(add_uuid, [bool], "add_uuid")
    core_utils.type_check(allow_lists, [bool], "allow_lists")

    if not allow_lists and isinstance(vector, (list, tuple)):
        raise ValueError("Lists are not allowed for vector.")

    vector_list = core_utils.ensure_list(vector)

    assert gdal_utils.is_vector_list(vector_list), f"Invalid vector in list: {vector_list}"

    path_list = gdal_utils.create_output_path_list(
        vector_list,
        out_path=out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _buffer_vector(
                in_vector,
                distance,
                out_path=path_list[index],
            )
        )

    if isinstance(vector, list):
        return output

    return output[0]



"""
### Bounding box utility functions ###

Various utility functions to work with bounding boxes and gdal.

There are two different formats for bounding boxes used by GDAL:</br>
OGR:  `[x_min, x_max, y_min, y_max]`</br>
WARP: `[x_min, y_min, x_max, y_max]`</br>

_If nothing else is stated, the OGR format is used._

The GDAL geotransform is a list of six parameters:</br>
`x_min, pixel_width, row_skew, y_max, column_skew, pixel_height (negative for north-up)`

TODO:
    * create tests
"""

# Standard library
import sys; sys.path.append("../../")
from uuid import uuid4

# External
import numpy as np
from osgeo import ogr, osr, gdal

# Internal
from buteo.utils import core_utils



def is_valid_bbox(bbox_ogr):
    """
    Checks if a bbox is valid.

    A valid ogr formatted bbox has the form: </br>
    `[x_min, x_max, y_min, y_max]`

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_bool_): **True** if the bbox is valid, **False** otherwise.
    """
    if not isinstance(bbox_ogr, list):
        return False

    if len(bbox_ogr) != 4:
        return False

    for val in bbox_ogr:
        if not core_utils.is_number(val):
            return False

    x_min, x_max, y_min, y_max = bbox_ogr

    if x_min > x_max or y_min > y_max:
        return False

    return True


def is_valid_bbox_latlng(bbox_ogr_latlng):
    """
    Checks if a bbox is valid and latlng.

    A valid ogr formatted bbox has the form: </br>
    `[x_min, x_max, y_min, y_max]`

    ## Args:
    `bbox_ogr_latlng` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_bool_): **True** if the bbox is valid, **False** otherwise.
    """
    if not is_valid_bbox(bbox_ogr_latlng):
        return False

    x_min, x_max, y_min, y_max = bbox_ogr_latlng
    if x_min < -180 or x_max > 180 or y_min < -90 or y_max > 90:
        return False

    return True


def is_valid_geotransform(geotransform):
    """
    Checks if a geotransform is valid.

    A valid geotransform has the form: </br>
    `[x_min, pixel_width, row_skew, y_max, column_skew, pixel_height]`

    ## Args:
    `geotransform` (_list_/_tuple_): A GDAL formatted geotransform.

    ## Returns:
    (_bool_): **True** if the geotransform is valid, **False** otherwise.
    """
    if not isinstance(geotransform, (list, tuple)):
        return False

    if len(geotransform) != 6:
        return False

    for val in geotransform:
        if not core_utils.is_number(val):
            return False

    return True


def ensure_negative(number):
    """
    Ensures that a valid is negative. If the number is positive, it is made negative.

    ## Args:
    `number` (_int_/_float_): A float or int number. </br>

    ## Returns:
    (_int_/_float_): The same number made **negative** if necesary.
    """
    assert core_utils.is_number(number), f"number must be a number. Received: {number}"

    if number <= 0:
        return -number

    return number


def get_pixel_offsets(geotransform, bbox_ogr):
    """
    Get the pixels offsets for a bbox and a geotransform.

    ## Args:
    `geotransform` (_list_/_tuple_): A GDAL GeoTransform. </br>
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_list_): A list of pixel offsets. `[x_start, y_start, x_size, y_size]`
    """
    assert is_valid_geotransform(
        geotransform
    ), f"geotransform must be a list of length six. Received: {geotransform}."
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr must be a valid OGR formatted bbox. Received: {bbox_ogr}."

    x_min, x_max, y_min, y_max = bbox_ogr

    origin_x = geotransform[0]
    origin_y = geotransform[3]
    pixel_width = abs(geotransform[1])
    pixel_height = abs(geotransform[5])

    x_1 = int((x_min - origin_x) / pixel_width)
    x_2 = int((x_max - origin_x) / pixel_width) + 1

    y_1 = int((y_max - origin_y) / pixel_height)
    y_2 = int((y_min - origin_y) / pixel_height) + 1

    x_size = x_2 - x_1
    y_size = y_2 - y_1

    x_start = x_1
    y_start = y_1

    return [x_start, y_start, x_size, y_size]


def get_bbox_from_geotransform(geotransform, raster_x_size, raster_y_size):
    """
    Get an OGR bounding box from a geotransform and raster sizes.

    ## Args:
    `geotransform` (_list_/_tuple_): A GDAL GeoTransform. </br>
    `raster_x_size` (_int_): The number of pixels in the x direction. </br>
    `raster_y_size` (_int_): The number of pixels in the y direction. </br>

    ## Returns:
    (_list_): An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert is_valid_geotransform(
        geotransform
    ), f"geotransform was not a valid geotransform. Received: {geotransform}"

    x_min, pixel_width, _row_skew, y_max, _column_skew, pixel_height = geotransform

    x_max = x_min + (raster_x_size * pixel_width)
    y_min = y_max + (raster_y_size * pixel_height)

    return [x_min, x_max, y_min, y_max]


def get_bbox_from_raster(raster_dataframe):
    """
    Gets an OGR bounding box from a GDAL raster dataframe.

    ## Args:
    `raster_dataframe` (_gdal.DataFrame_): A GDAL raster dataframe. </br>

    ## Returns:
    (_list_): An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        raster_dataframe, gdal.Dataset
    ), f"raster_dataframe was not a gdal.Datasource. Received: {raster_dataframe}"

    bbox = get_bbox_from_geotransform(
        raster_dataframe.GetGeoTransform(),
        raster_dataframe.RasterXSize,
        raster_dataframe.RasterYSize,
    )

    return bbox


def get_bbox_from_vector(vector_dataframe):
    """
    Gets an OGR bounding box from an OGR dataframe.

    ## Args:
    `vector_dataframe` (_ogr.DataSource_): An OGR vector dataframe. </br>

    ## Returns:
    (_list_): An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        vector_dataframe, ogr.DataSource
    ), f"vector_dataframe was not a valid ogr.DataSource. Received: {vector_dataframe}"

    layer_count = vector_dataframe.GetLayerCount()

    assert (
        layer_count > 0
    ), f"vector_dataframe did not contain any layers. Received: {vector_dataframe}"

    x_min = None
    x_max = None
    y_min = None
    y_max = None

    for layer_index in range(layer_count):
        layer = vector_dataframe.GetLayerByIndex(layer_index)
        layer_x_min, layer_x_max, layer_y_min, layer_y_max = layer.GetExtent()

        if layer_index == 0:
            x_min = layer_x_min
            x_max = layer_x_max
            y_min = layer_y_min
            y_max = layer_y_max
        else:
            if layer_x_min < x_min:
                x_min = layer_x_min
            if layer_x_max > x_max:
                x_max = layer_x_max
            if layer_y_min < y_min:
                y_min = layer_y_min
            if layer_y_max > y_max:
                y_max = layer_y_max

    return [x_min, x_max, y_min, y_max]


def get_bbox_from_vector_layer(vector_layer):
    """
    Gets an OGR bounding box from an OGR dataframe layer.

    ## Args:
    `vector_layer` (_ogr.Layer_): An OGR vector dataframe layer. </br>

    ## Returns:
    (_list_): An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        vector_layer, ogr.Layer
    ), f"vector_layer was not a valid ogr.Layer. Received: {vector_layer}"

    x_min, x_max, y_min, y_max = vector_layer.GetExtent()

    return [x_min, x_max, y_min, y_max]


def get_bbox_from_dataset(dataset):
    """
    Get the bbox from a dataset.

    ## Args:
    `dataset` (_str_/_gdal.Dataset_/_ogr.DataSource): A dataset or dataset path. </br>

    ## Returns:
    (_list_): The bounding box in ogr format: [x_min, x_max, y_min, y_max].
    """
    assert isinstance(dataset, (str, gdal.Dataset, ogr.DataSource)), "DataSet must be a string, ogr.DataSource, or gdal.Dataset."

    opened = dataset if isinstance(dataset, (gdal.Dataset, ogr.DataSource)) else None

    if opened is None:
        gdal.PushErrorHandler("CPLQuietErrorHandler")
        opened = gdal.Open(dataset, gdal.GA_ReadOnly)

        if opened is None:
            opened = ogr.Open(dataset, gdal.GA_ReadOnly)

        gdal.PopErrorHandler()
        if opened is None:
            raise Exception(f"Could not open dataset. {dataset}")

    if isinstance(opened, gdal.Dataset):
        return get_bbox_from_raster(opened)

    if isinstance(opened, ogr.DataSource):
        return get_bbox_from_vector(opened)

    raise Exception(f"Could not get bbox from dataset. {dataset}")


def get_sub_geotransform(geotransform, bbox_ogr):
    """
    Create a GeoTransform and the raster sizes for an OGR formatted bbox.

    ## Args:
    `geotransform` (_list_): A GDAL geotransform. </br>
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_dict_): { "Transform": _list_, "RasterXSize": _int_, "RasterYSize": _int_ }
    """
    assert is_valid_geotransform(
        geotransform
    ), f"geotransform must be a valid geotransform. Received: {geotransform}."
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr
    pixel_width = geotransform[1]
    pixel_height = geotransform[5]

    raster_x_size = round((x_max - x_min) / pixel_width)
    raster_y_size = round((y_max - y_min) / pixel_height)

    return {
        "Transform": [x_min, pixel_width, 0, y_max, 0, ensure_negative(pixel_height)],
        "RasterXSize": abs(raster_x_size),
        "RasterYSize": abs(raster_y_size),
    }


def convert_bbox_to_geom(bbox_ogr):
    """
    Convert an OGR bounding box to ogr.Geometry.</br>
    `[x_min, x_max, y_min, y_max] -> ogr.Geometry`

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_ogr.Geometry_): An OGR geometry.
    """
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(x_min, y_min)
    ring.AddPoint(x_max, y_min)
    ring.AddPoint(x_max, y_max)
    ring.AddPoint(x_min, y_max)
    ring.AddPoint(x_min, y_min)

    geom = ogr.Geometry(ogr.wkbPolygon)
    geom.AddGeometry(ring)

    return geom


def convert_geom_to_bbox(geom):
    """
    Convert an ogr.Geometry to an OGR bounding box.</br>
    `ogr.Geometry -> [x_min, x_max, y_min, y_max]`

    ## Args:
    `geom` (_ogr.Geometry_): An OGR geometry. </br>

    ## Returns:
    (_list_): An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        geom, ogr.Geometry
    ), f"geom was not a valid ogr.Geometry. Received: {geom}"

    bbox_ogr = list(geom.GetEnvelope()) # [x_min, x_max, y_min, y_max]

    return bbox_ogr


def convert_bbox_to_geotransform(bbox_ogr, raster_x_size, raster_y_size):
    """
    Convert an OGR formatted bounding box to a GDAL GeoTransform.</br>
    `[x_min, x_max, y_min, y_max] -> [x_min, pixel_width, x_skew, y_max, y_skew, pixel_height]`

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>
    `raster_x_size` (_int_): The number of pixels in the x direction. </br>
    `raster_y_size` (_int_): The number of pixels in the y direction. </br>

    ## Returns:
    (_list_): A GDAL GeoTransform. `[x_min, pixel_width, x_skew, y_max, y_skew, pixel_height]`
    """
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    origin_x = x_min
    origin_y = y_max
    pixel_width = (x_max - x_min) / raster_x_size
    pixel_height = (y_max - y_min) / raster_y_size

    return [origin_x, pixel_width, 0, origin_y, 0, ensure_negative(pixel_height)]


def convert_ogr_bbox_to_gdal_bbox(bbox_ogr):
    """
    Converts an OGR formatted bbox to a GDAL formatted one.</br>
    `[x_min, x_max, y_min, y_max] -> [x_min, y_min, x_max, y_max]`

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_list_): A GDAL formatted bbox. `[x_min, y_min, x_max, y_max]`
    """
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    return [x_min, y_min, x_max, y_max]


def convert_gdal_bbox_to_ogr_bbox(bbox_gdal):
    """
    Converts a GDAL formatted bbox to an OGR formatted one.</br>
    `[x_min, y_min, x_max, y_max] -> [x_min, x_max, y_min, y_max]`

    ## Args:
    `bbox_gdal` (_list_): A GDAL formatted bbox. </br>

    ## Returns:
    (_list_): An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert (
        isinstance(bbox_gdal, list) and len(bbox_gdal) == 4
    ), f"bbox_gdal must be a list of length four. Received: {bbox_gdal}."

    x_min, y_min, x_max, y_max = bbox_gdal

    return [x_min, x_max, y_min, y_max]


def convert_bbox_to_wkt(bbox_ogr):
    """
    Converts an OGR formatted bbox to a WKT string.</br>
    `[x_min, x_max, y_min, y_max] -> WKT`

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_str_): A WKT Polygon string. `POLYGON ((...))`
    """
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    wkt = f"POLYGON (({x_min} {y_min}, {x_max} {y_min}, {x_max} {y_max}, {x_min} {y_max}, {x_min} {y_min}))"

    return wkt


def convert_bbox_to_geojson(bbox_ogr):
    """
    Converts an OGR formatted bbox to a GeoJson dictionary.</br>
    `[x_min, x_max, y_min, y_max] -> GeoJson`

    ## Args:
    `bbox_ogr` (_list_): an OGR formatted bbox. </br>

    ## Returns:
    (_dict_): A GeoJson Dictionary. `{ "type": "Polygon", "coordinates": [ ... ] }`
    """
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    geojson = {
        "type": "Polygon",
        "coordinates": [
            [
                [x_min, y_min],
                [x_max, y_min],
                [x_max, y_max],
                [x_min, y_max],
                [x_min, y_min],
            ]
        ],
    }

    return geojson


def convert_bbox_to_vector(bbox_ogr, projection_osr):
    """
    Converts an OGR formatted bbox to an in-memory vector.</br>
    _Vectors are stored in /vsimem/ as .gpkg files._</br>
    **OBS**: Layers should be manually cleared when no longer used.

    ## Args:
    `bbox_ogr` (_list_): an OGR formatted bbox. </br>
    `projection_osr` (_osr.SpatialReference_): The projection of the vector. </br>

    ## Returns:
    (_ogr.DataSource_): The bounding box as a vector.
    """
    assert is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"
    assert isinstance(
        projection_osr, osr.SpatialReference
    ), f"projection_osr not a valid spatial reference. Recieved: {projection_osr}"

    geom = convert_bbox_to_geom(bbox_ogr)

    driver = ogr.GetDriverByName("GPKG")
    extent_name = f"/vsimem/{core_utils.get_unix_seconds_as_str()}_{uuid4().int}_extent.gpkg"

    extent_ds = driver.CreateDataSource(extent_name)

    layer = extent_ds.CreateLayer("extent_ogr", projection_osr, ogr.wkbPolygon)

    feature = ogr.Feature(layer.GetLayerDefn())
    feature.SetGeometry(geom)
    layer.CreateFeature(feature)

    feature = None
    layer.SyncToDisk()

    return extent_name


def bboxes_intersect(bbox1_ogr, bbox2_ogr):
    """
    Checks if two OGR formatted bboxes intersect.

    ## Args:
    `bbox1_ogr` (_list_): An OGR formatted bbox. </br>
    `bbox2_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_bool_): **True** if the bboxes intersect, **False** otherwise.
    """
    assert is_valid_bbox(
        bbox1_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox1_ogr}"
    assert is_valid_bbox(
        bbox2_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox2_ogr}"

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    if bbox2_x_min > bbox1_x_max:
        return False

    if bbox2_y_min > bbox1_y_max:
        return False

    if bbox2_x_max > bbox1_x_min:
        return False

    if bbox2_y_max > bbox1_y_min:
        return False

    return True


def bboxes_within(bbox1_ogr, bbox2_ogr):
    """
    Checks if one OGR formatted bbox is within another.

    ## Args:
    `bbox1_ogr` (_list_): An OGR formatted bbox. </br>
    `bbox2_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_bool_): **True** if the bbox is within the other, **False** otherwise.
    """
    assert is_valid_bbox(
        bbox1_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox1_ogr}"
    assert is_valid_bbox(
        bbox2_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox2_ogr}"

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    return (
        (bbox1_x_min >= bbox2_x_min)
        and (bbox1_x_max <= bbox2_x_max)
        and (bbox1_y_min >= bbox2_y_min)
        and (bbox1_y_max <= bbox2_y_max)
    )


def get_intersection_bboxes(bbox1_ogr, bbox2_ogr):
    """
    Get the intersection of two OGR formatted bboxes.

    ## Args:
    `bbox1_ogr` (_list_): An OGR formatted bbox. </br>
    `bbox2_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_list_): An OGR formatted bbox of the intersection. `[x_min, x_max, y_min, y_max]`
    """
    assert bboxes_intersect(
        bbox1_ogr, bbox2_ogr
    ), "The two bounding boxes do not intersect."

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    return [
        max(bbox1_x_min, bbox2_x_min),
        min(bbox1_x_max, bbox2_x_max),
        max(bbox1_y_min, bbox2_y_min),
        min(bbox1_y_max, bbox2_y_max),
    ]


def get_union_bboxes(bbox1_ogr, bbox2_ogr):
    """
    Get the union of two OGR formatted bboxes.

    ## Args:
    `bbox1_ogr` (_list_): An OGR formatted bbox. </br>
    `bbox2_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_list_): An OGR formatted bbox of the union. `[x_min, x_max, y_min, y_max]`
    """
    assert is_valid_bbox(
        bbox1_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox1_ogr}"
    assert is_valid_bbox(
        bbox2_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox2_ogr}"

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    return [
        min(bbox1_x_min, bbox2_x_min),
        max(bbox1_x_max, bbox2_x_max),
        min(bbox1_y_min, bbox2_y_min),
        max(bbox1_y_max, bbox2_y_max),
    ]


def get_projection_from_raster(raster):
    """
    Get the projection from a raster.

    ## Args:
    `raster` (_str_/_gdal.Dataset_): A raster or raster path. </br>

    ## Returns:
    (_osr.SpatialReference_): The projection in OSR format.
    """
    opened = None
    if isinstance(raster, gdal.Dataset):
        opened = raster
    else:
        gdal.PushErrorHandler("CPLQuietErrorHandler")
        opened = gdal.Open(raster, gdal.GA_ReadOnly)
        gdal.PopErrorHandler()

    if opened is None:
        raise Exception(f"Could not open raster. {raster}")

    projection = osr.SpatialReference()
    projection.ImportFromWkt(opened.GetProjection())
    opened = None

    return projection


def get_projection_from_vector(vector):
    """
    Get the projection from a vector.

    ## Args:
    `vector` (_str_/_gdal.Dataset_): A vector or vector path. </br>

    ## Returns:
    (_osr.SpatialReference_): The projection in OSR format.
    """
    opened = None
    if isinstance(vector, ogr.DataSource):
        opened = vector
    else:
        gdal.PushErrorHandler("CPLQuietErrorHandler")
        opened = ogr.Open(vector, gdal.GA_ReadOnly)
        gdal.PopErrorHandler()

    if opened is None:
        raise Exception(f"Could not open vector. {vector}")

    layer = opened.GetLayer()
    projection = layer.GetSpatialRef()
    opened = None

    return projection


def get_projection_from_dataset(dataset):
    """
    Get the projection from a dataset.

    ## Args:
    `dataset` (_str_/_gdal.Dataset_/_ogr.DataSource): A dataset or dataset path. </br>

    ## Returns:
    (_osr.SpatialReference_): The projection in OSR format.
    """
    assert isinstance(dataset, (str, gdal.Dataset, ogr.DataSource)), "DataSet must be a string, ogr.DataSource, or gdal.Dataset."

    opened = dataset if isinstance(dataset, (gdal.Dataset, ogr.DataSource)) else None

    if opened is None:
        gdal.PushErrorHandler("CPLQuietErrorHandler")
        opened = gdal.Open(dataset, gdal.GA_ReadOnly)

        if opened is None:
            opened = ogr.Open(dataset, gdal.GA_ReadOnly)

        gdal.PopErrorHandler()
        if opened is None:
            raise Exception(f"Could not open dataset. {dataset}")

    if isinstance(opened, gdal.Dataset):
        return get_projection_from_raster(opened)

    if isinstance(opened, ogr.DataSource):
        return get_projection_from_vector(opened)

    raise Exception(f"Could not get projection from dataset. {dataset}")


def align_bboxes_to_pixel_size(bbox1_ogr, bbox2_ogr, pixel_width, pixel_height):
    """
    Aligns two OGR formatted bboxes to a pixel size.

    ## Args:
    `bbox1_ogr` (_list_): An OGR formatted bbox. </br>
    `bbox2_ogr` (_list_): An OGR formatted bbox. </br>
    `pixel_width` (_float_/_int_): The width of the pixel. </br>
    `pixel_height` (_float_/_int_): The height of the pixel. </br>

    ## Returns:
    (_list_): An OGR formatted bbox of the alignment. `[x_min, x_max, y_min, y_max]`
    """
    assert (
        isinstance(bbox1_ogr, list) and len(bbox1_ogr) == 4
    ), f"bbox1_ogr must be a list of length four. Received: {bbox1_ogr}."
    assert (
        isinstance(bbox2_ogr, list) and len(bbox2_ogr) == 4
    ), f"bbox2_ogr must be a list of length four. Received: {bbox2_ogr}."

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    x_min = bbox2_x_min - ((bbox2_x_min - bbox1_x_min) % pixel_width)
    x_max = bbox2_x_max + ((bbox1_x_max - bbox2_x_max) % pixel_width)

    y_min = bbox2_y_min - ((bbox2_y_min - bbox1_y_min) % abs(pixel_height))
    y_max = bbox2_y_max + ((bbox1_y_max - bbox2_y_max) % abs(pixel_height))

    return x_min, x_max, y_min, y_max


def reproject_bbox(
    bbox_ogr,
    source_projection_osr,
    target_projection_osr,
):
    """
    Reprojects an OGR formatted bbox.

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>
    `source_projection_osr` (_osr.SpatialReference_): The source projection. </br>
    `target_projection_osr` (_osr.SpatialReference_): The target projection. </br>

    ## Returns:
    (_list_): An OGR formatted reprojected bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert is_valid_bbox(bbox_ogr), f"Invalid bbox. Received: {bbox_ogr}."
    assert isinstance(
        source_projection_osr, osr.SpatialReference
    ), f"source_projection not a valid spatial reference. Recieved: {source_projection_osr}"
    assert isinstance(
        target_projection_osr, osr.SpatialReference
    ), f"target_projection not a valid spatial reference. Recieved: {target_projection_osr}"

    if source_projection_osr.IsSame(target_projection_osr):
        return bbox_ogr

    transformer = osr.CoordinateTransformation(
        source_projection_osr, target_projection_osr
    )

    transformer = osr.CoordinateTransformation(source_projection_osr, target_projection_osr)

    x_min, x_max, y_min, y_max = bbox_ogr

    try:
        transformed_x_min, transformed_y_min, _z = transformer.TransformPoint(x_min, y_min)
        transformed_x_max, transformed_y_max, _z = transformer.TransformPoint(x_max, y_max)
    except: # pylint: disable=bare-except
        transformed_x_min, transformed_y_min, _z = transformer.TransformPoint(float(x_min), float(y_min))
        transformed_x_max, transformed_y_max, _z = transformer.TransformPoint(float(x_max), float(y_max))

    return [
        transformed_x_min,
        transformed_x_max,
        transformed_y_min,
        transformed_y_max,
    ]


def get_utm_zone_from_latlng(latlng, return_epsg=False):
    """ Get the UTM ZONE from a latlng list.

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_osr.SpatialReference_): The UTM zone projection.
    """
    assert isinstance(latlng, (list, np.ndarray)), "latlng must be in the form of a list."

    zone = round(((latlng[1] + 180) / 6) + 1)
    n_or_s = "S" if latlng[0] < 0 else "N"

    false_northing = "10000000" if n_or_s == "S" else "0"
    central_meridian = str(round(((zone * 6) - 180) - 3))
    epsg = f"32{'7' if n_or_s == 'S' else '6'}{str(zone)}"

    if return_epsg:
        return epsg

    wkt = f"""
        PROJCS["WGS 84 / UTM zone {str(zone)}{n_or_s}",
        GEOGCS["WGS 84",
            DATUM["WGS_1984",
                SPHEROID["WGS 84",6378137,298.257223563,
                    AUTHORITY["EPSG","7030"]],
                AUTHORITY["EPSG","6326"]],
            PRIMEM["Greenwich",0,
                AUTHORITY["EPSG","8901"]],
            UNIT["degree",0.0174532925199433,
                AUTHORITY["EPSG","9122"]],
            AUTHORITY["EPSG","4326"]],
        PROJECTION["Transverse_Mercator"],
        PARAMETER["latitude_of_origin",0],
        PARAMETER["central_meridian",{central_meridian}],
        PARAMETER["scale_factor",0.9996],
        PARAMETER["false_easting",500000],
        PARAMETER["false_northing",{false_northing}],
        UNIT["metre",1,
            AUTHORITY["EPSG","9001"]],
        AXIS["Easting",EAST],
        AXIS["Northing",NORTH],
        AUTHORITY["EPSG","{epsg}"]]
    """
    projection = osr.SpatialReference()
    projection.ImportFromWkt(wkt)

    return projection


def get_utm_zone_from_bbox(bbox_ogr_latlng):
    """
    Get the UTM zone from an OGR formatted bbox.

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>

    ## Returns:
    (_osr.SpatialReference_): The UTM zone projection.
    """
    assert is_valid_bbox(
        bbox_ogr_latlng
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr_latlng}"
    assert is_valid_bbox_latlng(bbox_ogr_latlng), "Bbox is not in latlng format."

    bbox_x_min, bbox_x_max, bbox_y_min, bbox_y_max = bbox_ogr_latlng

    mid_lng = (bbox_x_min + bbox_x_max) / 2
    mid_lat = (bbox_y_min + bbox_y_max) / 2

    return get_utm_zone_from_latlng([mid_lat, mid_lng])


def get_utm_zone_from_dataset(dataset):
    """
    Get the UTM zone from a GDAL dataset.

    ## Args:
    `dataset` (_obj_): A GDAL dataset. </br>

    ## Returns:
    (_str_): The UTM zone.
    """
    assert isinstance(dataset, (str, gdal.Dataset, ogr.DataSource)), "dataset was not a valid dataset."

    bbox = get_bbox_from_dataset(dataset)
    source_projection = get_projection_from_dataset(dataset)
    target_projection = osr.SpatialReference()
    # target_projection.ImportFromEPSG(4326)
    target_projection.ImportFromWkt('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]')

    bbox = reproject_bbox(bbox, source_projection, target_projection)
    utm_zone = get_utm_zone_from_bbox(bbox)

    return utm_zone


def get_utm_zone_from_dataset_list(datasets):
    """
    Get the UTM zone from a list of GDAL datasets.

    ## Args:
    `datasets` (_list_): A list of GDAL datasets. </br>

    ## Returns:
    (_str_): The UTM zone.
    """
    assert isinstance(datasets, (list, np.ndarray)), "datasets was not a valid list."

    latlng_bboxes = []
    latlng_proj = osr.SpatialReference()
    # latlng_proj.ImportFromEPSG(4326)
    latlng_proj.ImportFromWkt('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]')

    for dataset in datasets:
        bbox = get_bbox_from_dataset(dataset)
        projection = get_projection_from_dataset(dataset)
        latlng_bboxes.append(reproject_bbox(bbox, projection, latlng_proj))

    union_bbox = latlng_bboxes[0]
    for bbox in latlng_bboxes[1:]:
        union_bbox = get_union_bboxes(union_bbox, bbox)

    utm_zone = get_utm_zone_from_bbox(union_bbox)

    return utm_zone


def reproject_latlng_point_to_utm(latlng):
    """ Converts a latlng point into an UTM point.

        Takes point in [lat, lng], returns [utm_x, utm_y].
    """
    source_projection = osr.SpatialReference()
    # source_projection.ImportFromEPSG(4326)
    source_projection.ImportFromWkt('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]')
    target_projection = get_utm_zone_from_latlng(latlng)

    transformer = osr.CoordinateTransformation(
        source_projection, target_projection
    )

    try:
        utm_x, utm_y, _utm_z = transformer.TransformPoint(latlng[0], latlng[1])
    except: # pylint: disable=bare-except
        utm_x, utm_y, _utm_z = transformer.TransformPoint(float(latlng[0]), float(latlng[1]))

    return [utm_x, utm_y]


def additional_bboxes(bbox_ogr, projection_osr):
    """
    This is an internal utility function for metadata generation. It takes a standard
    OGR bounding box and returns a list of variations of bounding boxes.

    ## Args:
    `bbox_ogr` (_list_): An OGR formatted bbox. </br>
    `projection_osr` (_osr.SpatialReference_): The projection. </br>

    ## Returns:
    (_dict_): A dictionary of the added bboxes. Contains the following keys: </br>
    `bbox_latlng`: The bbox in latlng coordinates. </br>
    `bbox_wkt`: The bbox in WKT format. </br>
    `bbox_wkt_latlng`: The bbox in WKT format in latlng coordinates. </br>
    `bbox_geom`: The bbox in ogr.Geometry format. </br>
    `bbox_geom_latlng`: The bbox in ogr.Geometry format in latlng coordinates. </br>
    `bbox_gdal`: The bbox in GDAL format. </br>
    `bbox_gdal_latlng`: The bbox in GDAL format in latlng coordinates. </br>
    `bbox_dict`: The bbox in a dictionary format. { "x_min": x_min, ... } </br>
    `bbox_dict_latlng`: The bbox in a dictionary format in latlng coordinates. </br>
    """
    assert is_valid_bbox(bbox_ogr), f"Invalid bbox. Received: {bbox_ogr}."
    assert isinstance(
        projection_osr, osr.SpatialReference
    ), f"source_projection not a valid spatial reference. Recieved: {projection_osr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    original_projection = osr.SpatialReference()
    original_projection.ImportFromWkt(projection_osr.ExportToWkt())

    latlng_projection = osr.SpatialReference()
    # latlng_projection.ImportFromEPSG(4326)
    latlng_projection.ImportFromWkt('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]')

    bbox_ogr_latlng = reproject_bbox(bbox_ogr, original_projection, latlng_projection)
    latlng_x_min, latlng_x_max, latlng_y_min, latlng_y_max = bbox_ogr_latlng

    bbox_geom = convert_bbox_to_geom(bbox_ogr)
    bbox_geom_latlng = convert_bbox_to_geom(bbox_ogr_latlng)

    bbox_wkt = convert_bbox_to_wkt(bbox_ogr)
    bbox_wkt_latlng = convert_bbox_to_wkt(bbox_ogr_latlng)

    return {
        "bbox_latlng": bbox_ogr_latlng,
        "bbox_wkt": bbox_wkt,
        "bbox_wkt_latlng": bbox_wkt_latlng,
        "bbox_geom": bbox_geom,
        "bbox_geom_latlng": bbox_geom_latlng,
        "bbox_gdal": convert_ogr_bbox_to_gdal_bbox(bbox_ogr),
        "bbox_gdal_latlng": convert_ogr_bbox_to_gdal_bbox(bbox_ogr_latlng),
        "bbox_dict": {"x_min": x_min, "x_max": x_max, "y_min": y_min, "y_max": y_max},
        "bbox_dict_latlng": {
            "x_min": latlng_x_min,
            "x_max": latlng_x_max,
            "y_min": latlng_y_min,
            "y_max": latlng_y_max,
        },
        "bbox_geojson": convert_bbox_to_geojson(bbox_ogr_latlng),
    }



# """
# Calculate zonal statistics from vector and raster files.

# BROKEN

# TODO:
#     - Fix rasterize step
#     - Improve documentation

# """

# import sys; sys.path.append("../../") # Path: buteo/vector/zonal_statistics.py
# import numpy as np

# from numba import jit
# from osgeo import ogr

# from buteo.raster.core_raster import raster_to_array, raster_to_metadata
# from buteo.vector.rasterize import rasterize_vector
# from buteo.vector.reproject import reproject_vector
# from buteo.vector.core_vector import (
#     open_vector,
#     _vector_to_memory,
#     _vector_to_metadata,
# )
# from buteo.vector.zonal_statistics_stats import calculate_array_stats
# from buteo.utils.core_utils import progress



# import numpy as np
# from enum import Enum

# from numba import jit


# class stat(Enum):
#     count = 1
#     range = 2
#     min = 3
#     max = 4
#     sum = 5
#     mean = 6
#     avg = 6
#     average = 6
#     var = 7
#     variance = 7
#     std = 8
#     stdev = 8
#     standard_deviation = 8
#     skew = 9
#     kurtosis = 10
#     median = 11
#     med = 11
#     iqr = 12
#     q02 = 13
#     q98 = 14
#     q1 = 15
#     q3 = 16
#     mad = 17
#     median_absolute_deviation = 17
#     mode = 18
#     snr = 19
#     eff = 20
#     cv = 21


# @jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
# def calculate_array_stats(arr, stats):
#     stats_length = int(len(stats))
#     result = np.zeros((stats_length), dtype="float32")

#     for idx in range(stats_length):
#         if stat(stats[idx]) == 1:  # count
#             result[idx] = arr.size
#         elif stat(stats[idx]) == 2:  # range
#             result[idx] = np.ptp(arr)
#         elif stat(stats[idx]) == 3:  # min
#             result[idx] = np.min(arr)
#         elif stat(stats[idx]) == 4:  # max
#             result[idx] = np.max(arr)
#         elif stat(stats[idx]) == 5:  # sum
#             result[idx] = np.sum(arr)
#         elif stat(stats[idx]) == 6:  # mean
#             result[idx] = np.mean(arr)
#         elif stat(stats[idx]) == 7:  # var
#             result[idx] = np.var(arr)
#         elif stat(stats[idx]) == 8:  # std
#             result[idx] = np.std(arr)
#         elif stat(stats[idx]) == 9:  # skew
#             mean = np.mean(arr)
#             std = np.std(arr)
#             if std == 0:
#                 result[idx] = 0.0
#                 continue
#             deviations = np.sum(np.power(arr - mean, 3))
#             result[idx] = (deviations * (1 / arr.size)) / (np.power(std, 3))
#         elif stat(stats[idx]) == 10:  # kurt
#             mean = np.mean(arr)
#             std = np.std(arr)
#             if std == 0:
#                 result[idx] = 0.0
#                 continue
#             deviations = np.sum(np.power(arr - mean, 4))
#             result[idx] = (deviations * (1 / arr.size)) / (np.power(std, 4))
#         elif stat(stats[idx]) == 11:  # median
#             result[idx] = np.median(arr)
#         elif stat(stats[idx]) == 12:  # iqr
#             result[idx] = np.quantile(
#                 arr, np.array([0.25, 0.75], dtype="float32")
#             ).sum()
#         elif stat(stats[idx]) == 13:  # q02
#             result[idx] = np.quantile(arr, 0.02)
#         elif stat(stats[idx]) == 14:  # q92
#             result[idx] = np.quantile(arr, 0.98)
#         elif stat(stats[idx]) == 15:  # q1
#             result[idx] = np.quantile(arr, 0.25)
#         elif stat(stats[idx]) == 16:  # q3
#             result[idx] = np.quantile(arr, 0.75)
#         elif stat(stats[idx])== 17:  # mad
#             median = np.median(arr)
#             absdev = np.abs(arr - median)
#             result[idx] = np.median(absdev)
#         elif stat(stats[idx]) == 18:  # mode
#             uniques = np.unique(arr)
#             counts = np.zeros_like(uniques, dtype="uint64")

#             # numba does not support return count of uniques.
#             for idx_values in range(len(arr)):
#                 val = arr[idx_values]
#                 for idx_uniques in range(len(uniques)):
#                     if val == uniques[idx_uniques]:
#                         counts[idx_uniques] = counts[idx_uniques] + 1
#             index = np.argmax(counts)
#             result[idx] = uniques[index]
#         elif stat(stats[idx]) == 19:  # snr
#             std = np.std(arr)
#             if std == 0:
#                 result[idx] = 0.0
#                 continue
#             result[idx] = np.mean(arr) / std
#         elif stat(stats[idx]) == 20:  # eff
#             mean = np.mean(arr)
#             if mean == 0:
#                 result[idx] = 0.0
#                 continue
#             result[idx] = np.var(arr) / np.power(mean, 2)
#         elif stat(stats[idx]) == 21:  # cv
#             mean = np.mean(arr)
#             if mean == 0:
#                 result[idx] = 0.0
#                 continue
#             result[idx] = np.std(arr) / mean

#     return result

# @jit(nopython=True, nogil=True, fastmath=True, inline="always")
# def overlap_size_calc(extent, raster_transform):
#     return np.array(
#         [
#             (extent[1] - extent[0]) / raster_transform[1],
#             (extent[3] - extent[2]) / abs(raster_transform[5]),
#         ],
#         dtype=np.int32,
#     )


# @jit(nopython=True, nogil=True, fastmath=True, inline="always")
# def align_extent(raster_transform, vector_extent, raster_size):
#     pixel_width = abs(raster_transform[1])
#     pixel_height = abs(raster_transform[5])

#     raster_min_x = raster_transform[0]
#     # raster_max_x = raster_min_x + (raster_size[0] * pixel_width)
#     raster_max_y = raster_transform[3]
#     raster_min_y = raster_max_y + (raster_size[1] * -pixel_width)

#     vector_min_x = vector_extent[0]
#     vector_max_x = vector_extent[1]
#     vector_min_y = vector_extent[2]
#     vector_max_y = vector_extent[3]

#     # Align the two extents
#     vector_min_x = vector_min_x - (vector_min_x - raster_min_x) % pixel_width
#     vector_max_x = vector_max_x + (vector_max_x - raster_min_x) % pixel_width
#     vector_min_y = vector_min_y - (vector_min_y - raster_max_y) % pixel_height
#     vector_max_y = vector_max_y + (vector_max_y - raster_max_y) % pixel_height

#     rasterized_x_size = int((vector_max_x - vector_min_x) / pixel_width)
#     rasterized_y_size = int((vector_max_y - vector_min_y) / pixel_height)

#     rasterized_x_offset = int((vector_min_x - raster_min_x) / pixel_width)
#     rasterized_y_offset = (
#         int(raster_size[1] - int((vector_min_y - raster_min_y) / pixel_height))
#         - rasterized_y_size
#     )

#     if rasterized_x_offset < 0:
#         rasterized_x_offset = 0

#     if rasterized_y_offset < 0:
#         rasterized_y_offset = 0

#     if (rasterized_x_offset + rasterized_x_size) > raster_size[0]:
#         rasterized_x_offset = rasterized_x_offset - (
#             (rasterized_x_offset + rasterized_x_size) - raster_size[0]
#         )

#     if (rasterized_y_offset + rasterized_y_size) > raster_size[1]:
#         rasterized_y_offset = rasterized_y_offset - (
#             (rasterized_y_offset + rasterized_y_size) - raster_size[1]
#         )

#     new_vector_extent = np.array(
#         [vector_min_x, vector_max_x, vector_min_y, vector_max_y], dtype=np.float32
#     )
#     rasterized_size = np.array(
#         [rasterized_x_size, rasterized_y_size, pixel_width, pixel_height],
#         dtype=np.float32,
#     )
#     offset = np.array([rasterized_x_offset, rasterized_y_offset], dtype=np.int32)

#     return new_vector_extent, rasterized_size, offset


# @jit(nopython=True, nogil=True, fastmath=True, inline="always")
# def get_intersection(extent1, extent2):
#     one_bottomLeftX = extent1[0]
#     one_topRightX = extent1[1]
#     one_bottomLeftY = extent1[2]
#     one_topRightY = extent1[3]

#     two_bottomLeftX = extent2[0]
#     two_topRightX = extent2[1]
#     two_bottomLeftY = extent2[2]
#     two_topRightY = extent2[3]

#     if two_bottomLeftX > one_topRightX:  # Too far east
#         return np.array([0, 0, 0, 0], dtype=np.float32)
#     elif two_bottomLeftY > one_topRightY:  # Too far north
#         return np.array([0, 0, 0, 0], dtype=np.float32)
#     elif two_topRightX < one_bottomLeftX:  # Too far west
#         return np.array([0, 0, 0, 0], dtype=np.float32)
#     elif two_topRightY < one_bottomLeftY:  # Too far south
#         return np.array([0, 0, 0, 0], dtype=np.float32)
#     else:

#         x_min = (
#             one_bottomLeftX if one_bottomLeftX > two_bottomLeftX else two_bottomLeftX
#         )
#         x_max = one_topRightX if one_topRightX < two_topRightX else two_topRightX
#         y_min = (
#             one_bottomLeftY if one_bottomLeftY > two_bottomLeftY else two_bottomLeftY
#         )
#         y_max = one_topRightY if one_topRightY < two_topRightY else two_topRightY

#         return np.array([x_min, x_max, y_min, y_max], dtype=np.float32)


# @jit(nopython=True, nogil=True, fastmath=True, inline="always")
# def get_extent(raster_transform, raster_size):
#     bottomRightX = raster_transform[0] + (raster_size[0] * raster_transform[1])
#     bottomRightY = raster_transform[3] + (raster_size[1] * raster_transform[5])

#     return np.array(
#         [raster_transform[0], bottomRightX, bottomRightY, raster_transform[3]],
#         dtype=np.float32,
#     )


# def zonal_statistics(
#     in_vector,
#     output_vector=None,
#     in_rasters=[],
#     prefixes=[],
#     stats=None,
# ):
#     """
#     ZonalStatistics.
#     """
#     default_stats = ["mean", "med", "std"]

#     if stats is None:
#         stats = default_stats

#     if len(prefixes) != 0:
#         if len(in_rasters) != len(prefixes):
#             raise ValueError("Unable to parse prefixes.")

#     if isinstance(in_rasters, list):
#         if len(in_rasters) == 0:
#             raise ValueError("List of rasters (in_rasters) is empty.")

#     if len(stats) == 0:
#         raise ValueError("Unable to parse statistics (stats).")

#     # Read the raster meta:
#     raster_metadata = raster_to_metadata(in_rasters[0])

#     vector = None
#     if output_vector is None:
#         vector = open_vector(in_vector, writeable=True)
#     else:
#         vector = _vector_to_memory(in_vector)

#     vector_metadata = _vector_to_metadata(vector)
#     vector_layer = vector.GetLayer()

#     # Check that projections match
#     if not vector_metadata["projection_osr"].IsSame(raster_metadata["projection_osr"]):
#         if output_vector is None:
#             vector = reproject_vector(in_vector, in_rasters[0])
#         else:
#             vector_path = reproject_vector(
#                 in_vector, in_rasters[0], output_vector
#             )
#             vector = open_vector(vector_path, writeable=True)

#         vector_metadata = _vector_to_metadata(vector)
#         vector_layer = vector.GetLayer()

#     vector_projection = vector_metadata["projection_osr"]
#     raster_projection = raster_metadata["projection_osr"]

#     # Read raster data in overlap
#     raster_transform = np.array(raster_metadata["transform"], dtype=np.float32)
#     raster_size = np.array(raster_metadata["size"], dtype=np.int32)

#     raster_extent = get_extent(raster_transform, raster_size)

#     vector_extent = np.array(vector_layer.GetExtent(), dtype=np.float32)
#     overlap_extent = get_intersection(raster_extent, vector_extent)

#     if overlap_extent is False:
#         print("raster_extent: ", raster_extent)
#         print("vector_extent: ", vector_extent)
#         raise Exception("Vector and raster do not overlap!")

#     (
#         overlap_aligned_extent,
#         overlap_aligned_rasterized_size,
#         overlap_aligned_offset,
#     ) = align_extent(raster_transform, overlap_extent, raster_size)
#     overlap_transform = np.array(
#         [
#             overlap_aligned_extent[0],
#             raster_transform[1],
#             0,
#             overlap_aligned_extent[3],
#             0,
#             raster_transform[5],
#         ],
#         dtype=np.float32,
#     )
#     overlap_size = overlap_size_calc(overlap_aligned_extent, raster_transform)

#     # Loop the features
#     vector_driver = ogr.GetDriverByName("Memory")
#     vector_feature_count = vector_layer.GetFeatureCount()
#     vector_layer.StartTransaction()

#     # Create fields
#     vector_layer_defn = vector_layer.GetLayerDefn()
#     vector_field_counts = vector_layer_defn.GetFieldCount()
#     vector_current_fields = []

#     # Get current fields
#     for i in range(vector_field_counts):
#         vector_current_fields.append(vector_layer_defn.GetFieldDefn(i).GetName())

#     # Add fields where missing
#     for stat in stats:
#         for i in range(len(in_rasters)):
#             field_name = f"{prefixes[i]}{stat}"
#             if field_name not in vector_current_fields:
#                 field_defn = ogr.FieldDefn(field_name, ogr.OFTReal)
#                 vector_layer.CreateField(field_defn)

#     rasterized_features = []
#     sizes = np.zeros((vector_feature_count, 4), dtype="float32")
#     offsets = np.zeros((vector_feature_count, 2), dtype=np.int32)
#     raster_data = None
#     for raster_index, raster_value in enumerate(in_rasters):

#         columns = {}
#         for stat in stats:
#             columns[prefixes[raster_index] + stat] = []

#         fits_in_memory = True
#         try:
#             raster_data = raster_to_array(
#                 raster_value,
#                 pixel_offsets=[
#                     overlap_aligned_offset[0],
#                     overlap_aligned_offset[1],
#                     overlap_aligned_rasterized_size[0],
#                     overlap_aligned_rasterized_size[1],
#                 ],
#             )
#         except:
#             fits_in_memory = False
#             print("Raster does not fit in memory.. Doing IO for each feature.")

#         for n in range(vector_feature_count):
#             vector_feature = vector_layer.GetNextFeature()
#             rasterized_vector = None

#             if raster_index == 0:

#                 try:
#                     vector_geom = vector_feature.GetGeometryRef()
#                 except:
#                     vector_geom.Buffer(0)
#                     Warning("Invalid geometry at : ", n)

#                 if vector_geom is None:
#                     raise Exception("Invalid geometry. Could not fix.")

#                 feature_extent = vector_geom.GetEnvelope()

#                 # Create temp layer
#                 temp_vector_datasource = vector_driver.CreateDataSource(f"vector_{n}")
#                 temp_vector_layer = temp_vector_datasource.CreateLayer(
#                     "temp_polygon", vector_projection, ogr.wkbPolygon
#                 )
#                 temp_vector_layer.CreateFeature(vector_feature.Clone())

#                 (
#                     feature_aligned_extent,
#                     feature_aligned_rasterized_size,
#                     feature_aligned_offset,
#                 ) = align_extent(overlap_transform, feature_extent, overlap_size)

#                 rasterized_vector = None
#                 # rasterized_vector = rasterize_vector(
#                 #     temp_vector_layer,
#                 #     feature_aligned_extent,
#                 #     feature_aligned_rasterized_size,
#                 #     raster_projection,
#                 # )
#                 rasterized_features.append(rasterized_vector)

#                 offsets[n] = feature_aligned_offset
#                 sizes[n] = feature_aligned_rasterized_size

#             if fits_in_memory is True:
#                 cropped_raster = raster_data[
#                     offsets[n][1] : offsets[n][1] + int(sizes[n][1]),  # X
#                     offsets[n][0] : offsets[n][0] + int(sizes[n][0]),  # Y
#                 ]
#             else:
#                 cropped_raster = raster_to_array(
#                     raster_value,
#                     pixel_offsets=[
#                         overlap_aligned_offset[0] + offsets[n][0],
#                         overlap_aligned_offset[1] + offsets[n][1],
#                         int(sizes[n][0]),
#                         int(sizes[n][1]),
#                     ],
#                 )

#             if rasterized_features[n] is None:
#                 for stat in stats:
#                     field_name = f"{prefixes[raster_index]}{stat}"
#                     vector_feature.SetField(field_name, None)
#             elif cropped_raster is None:
#                 for stat in stats:
#                     field_name = f"{prefixes[raster_index]}{stat}"
#                     vector_feature.SetField(field_name, None)
#             else:
#                 raster_data_masked = np.ma.masked_array(
#                     cropped_raster, mask=rasterized_features[n], dtype="float32"
#                 ).compressed()
#                 zonal_stats = calculate_array_stats(
#                     raster_data_masked, stats
#                 )

#                 for index, stat in enumerate(stats):
#                     field_name = f"{prefixes[raster_index]}{stat}"
#                     vector_feature.SetField(field_name, float(zonal_stats[index]))

#                 vector_layer.SetFeature(vector_feature)

#             progress(n, vector_feature_count, name=prefixes[raster_index])

#         vector_layer.ResetReading()

#     vector_layer.CommitTransaction()

#     if output_vector is None:
#         return vector

#     return output_vector



"""
### Resample rasters. ###

Module to resample rasters to a target resolution.
Can uses references from vector or other raster datasets.
"""

# Standard library
import sys; sys.path.append("../../")

# External
from osgeo import gdal
import numpy as np

# Internal
from buteo.utils import core_utils, gdal_utils, gdal_enums
from buteo.raster import core_raster



def _resample_raster(
    raster,
    target_size,
    out_path=None,
    *,
    target_in_pixels=False,
    resample_alg="nearest",
    overwrite=True,
    creation_options=None,
    dtype=None,
    dst_nodata="infer",
    prefix="",
    suffix="_resampled",
    add_uuid=False,
):
    """ Internal. """
    assert isinstance(raster, (gdal.Dataset, str)), f"The input raster must be in the form of a str or a gdal.Dataset: {raster}"

    out_path = gdal_utils.create_output_path(
        raster,
        out_path=out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    ref = core_raster._open_raster(raster)
    metadata = core_raster._raster_to_metadata(ref)

    x_res, y_res, x_pixels, y_pixels = gdal_utils.parse_raster_size(
        target_size, target_in_pixels=target_in_pixels
    )

    out_format = gdal_utils.path_to_driver_raster(out_path)

    src_nodata = metadata["nodata_value"]
    out_nodata = None
    if dst_nodata == "infer":
        dst_nodata = src_nodata
    else:
        assert isinstance(dst_nodata, (int, float, type(None))), "dst_nodata must be an int, float, 'infer', or 'None'"
        out_nodata = dst_nodata

    if dtype is None:
        dtype = metadata["datatype"]

    core_utils.remove_if_required(out_path, overwrite)

    resampled = gdal.Warp(
        out_path,
        ref,
        width=x_pixels,
        height=y_pixels,
        xRes=x_res,
        yRes=y_res,
        format=out_format,
        outputType=gdal_enums.translate_str_to_gdal_dtype(dtype),
        resampleAlg=gdal_enums.translate_resample_method(resample_alg),
        creationOptions=gdal_utils.default_creation_options(creation_options),
        srcNodata=metadata["nodata_value"],
        dstNodata=out_nodata,
        multithread=True,
    )

    if resampled is None:
        raise Exception(f"Error while resampling raster: {out_path}")

    return out_path


def resample_raster(
    raster,
    target_size,
    out_path=None,
    *,
    target_in_pixels=False,
    resample_alg="nearest",
    creation_options=None,
    dtype=None,
    dst_nodata="infer",
    prefix="",
    suffix="",
    add_uuid=False,
    overwrite=True,
):
    """
    Reprojects raster(s) given a target projection. </br>
    **Beware** if your input is in latitude and longitude, you'll need to specify the target_size in degrees as well.

    ## Args:
    `raster` (_str_/_list_/_gdal.Dataset): The raster(s) to reproject. </br>
    `target_size` (_str_/_int_/_ogr.DataSource_/_gdal.Dataset_): The target resolution of the
    raster. In the same unit as the projection of the raster.
    It's better to reproject to a projected coordinate system for resampling.
    If a raster is the target_size the function will read the pixel size from
    that raster. </br>

    ## Kwargs:
    `target_in_pixels` (_bool_): If True, the target_size will be interpreted as the number of pixels. (Default: **False**) </br>
    `out_path` (_str_/_None_): The output path. If not provided, the output path is inferred from the input. (Default: **None**) </br>
    `resample_alg` (_str_): The resampling algorithm. (Default: **nearest**) </br>
    `copy_if_same` (_bool_): If the input and output projections are the same, copy the input raster to the output path. (Default: **True**) </br>
    `overwrite` (_bool_): If the output path already exists, overwrite it. (Default: **True**) </br>
    `creation_options` (_list_/_None_): A list of creation options for the output raster. (Default: **None**) </br>
    `dst_nodata` (_str_/_int_/_float_): The nodata value for the output raster. (Default: **infer**) </br>
    `prefix` (_str_): The prefix to add to the output path. (Default: **""**) </br>
    `suffix` (_str_): The suffix to add to the output path. (Default: **"_reprojected"**) </br>
    `add_uuid` (_bool_): If True, add a UUID to the output path. (Default: **False**) </br>

    ## Returns:
    (_str_/_list_): The output path(s) of the resampled raster(s).
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    core_utils.type_check(target_size, [tuple, [int, float], int, float, str, gdal.Dataset], "target_size")
    core_utils.type_check(target_in_pixels, [bool], "target_in_pixels")
    core_utils.type_check(out_path, [str, [str], None], "out_path")
    core_utils.type_check(resample_alg, [str], "resample_alg")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(creation_options, [[str], None], "creation_options")
    core_utils.type_check(dst_nodata, [str, int, float, None], "dst_nodata")
    core_utils.type_check(dtype, [str, None], "dtype")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "postfix")

    if core_utils.is_str_a_glob(raster):
        raster = core_utils.parse_glob_path(raster)

    raster_list = core_utils.ensure_list(raster)
    assert gdal_utils.is_raster_list(raster_list), f"Invalid raster in raster list: {raster_list}"

    path_list = gdal_utils.create_output_path_list(
        raster_list,
        out_path=out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    resampled_rasters = []
    for index, in_raster in enumerate(raster_list):
        resampled_rasters.append(
            _resample_raster(
                in_raster,
                target_size,
                target_in_pixels=target_in_pixels,
                out_path=path_list[index],
                resample_alg=resample_alg,
                overwrite=overwrite,
                creation_options=creation_options,
                dtype=dtype,
                dst_nodata=dst_nodata,
                prefix=prefix,
                suffix=suffix,
            )
        )

    if isinstance(raster, list):
        return resampled_rasters

    return resampled_rasters[0]


def resample_array(arr, target_shape_pixels, resample_alg="nearest"):
    """ Resample a numpy array using the GDAL algorithms. """
    core_utils.type_check(arr, [np.ndarray, np.ma.MaskedArray], "arr")
    core_utils.type_check(target_shape_pixels, [tuple, [int, float]], "target_shape_pixels")
    core_utils.type_check(resample_alg, [str], "resample_alg")

    if len(target_shape_pixels) > 2:
        target_shape_pixels = target_shape_pixels[:2]

    arr_as_raster = core_raster.create_raster_from_array(arr)
    resampled = _resample_raster(arr_as_raster, target_shape_pixels, target_in_pixels=True, resample_alg=resample_alg)
    out_arr = core_raster.raster_to_array(resampled)

    gdal_utils.delete_if_in_memory(arr_as_raster)
    gdal_utils.delete_if_in_memory(resampled)

    return out_arr



"""
Make simple work of raster analysis!
"""
from .core_raster import *
from .clip import *
from .align import *
from .borders import *
from .grid import *
from .nodata import *
from .proximity import *
from .reproject import *
from .resample import *
from .shift import *
from .vectorize import *
from .warp import *
from .convolution import *
from .edge_detection import *
from .morphology import *
